@@ -36,11 +36,14 @@
 try:
     # # %load_ext Cython
     # !python3 setup.py build_ext --inplace
-    # from .cagg import aggregate
-    from .aggregation_memview import aggregate # cython with memory view
+    from .aggregation_cm import aggregate # cython with memory view
+    from .amalgamation_cm import *
 except ModuleNotFoundError:
     from .aggregation import aggregate
+    from .amalgamation import *
+    print("Cython fail.")
 
+    
 import os
 import copy
 import warnings
@@ -52,6 +55,9 @@ from numpy.linalg import norm
 #import pyximport; pyximport.install()
 from matplotlib import pyplot as plt
 from sklearn.decomposition import PCA
+from scipy.sparse import csr_matrix
+
+
 np.random.seed(0)
 # from scipy.sparse import csr_matrix
 # from scipy.sparse.csgraph import shortest_path
@@ -61,7 +67,7 @@ np.random.seed(0)
 
 # ##########################################################################################################
 # ******************************************** the main wrapper ********************************************
-class classix:
+class CLASSIX:
     """CLASSIX: Fast and explainable clustering based on sorting.
 
     Parameters
@@ -491,72 +497,16 @@ class classix:
         # labels = np.array(labels)
         labels = copy.deepcopy(agg_labels) 
         
-        
         # calculate the overall volume of the combination of the clusters,
         # hence only compare their neighbor clusters
         # print("clusters merging initialize...")
-        
-        self.connected_pairs = list()
-        for i in tqdm(splist[:, 1], disable=not self.verbose):
-            sp1 = splist[int(i), 3:]
-
-            if method == "density":                    # calculate the density
-                volume = np.pi**(data.shape[1]/2) * radius**data.shape[1] / gamma(data.shape[1]/2+1)
-                index_overlap = np.where(np.linalg.norm(splist[:,3:] - sp1, ord=2, axis=-1) <= 2*radius)  # 2*distance_scale*radius
-                # index_overlap = np.where(np.linalg.norm(splist[:,3:] - sp1, ord=2, axis=-1) <= 2*radius)  # -> [for np.where, removed], 2*distance_scale*radius
-
-                # calculate their neighbors ...1)
-                # later decide whether sps should merge with neighbors
-                if not np.any(index_overlap):
-                    continue
-
-                # if len(index_overlap[0]) <= 0:  # -> [for np.where, removed],  make sure the neighbors is not empty
-                #     continue
-
-                neigbor_sp = splist[index_overlap] # calculate their neighbor ...2)
-                neigbor_sp = neigbor_sp[neigbor_sp[:,1] > i]
-                
-                c1 = np.linalg.norm(data-sp1, ord=2, axis=-1) <= 2*radius
-                den1 = np.count_nonzero(c1) / volume
-                # den1 = splist[int(i), 2] / volume # density(splist[int(i), 2], volume = volume) 
-                for j in neigbor_sp[:,1]:
-                    sp2 = splist[int(j), 3:]
-                    
-                    c2 = np.linalg.norm(data-sp2, ord=2, axis=-1) <= 2*radius
-                    den2 = np.count_nonzero(c2) / volume
-                    #den2 = splist[int(j), 2] / volume # density(splist[int(j), 2], volume = volume)
-                    if check_if_overlap(sp1, sp2, radius=radius): # , scale=distance_scale
-                        # interdata = np.linalg.norm(data[agg_labels==i] - sp2, ord=2, axis=-1)
-                        # internum = len(interdata[interdata <= 2*radius])
-                        internum = np.count_nonzero(c1 & c2)
-                        cid = cal_inter_density(sp1, sp2, radius=radius, num=internum)
-
-                        if cid >= den1 or cid >= den2: # eta*cid >= den1 or eta*cid >= den2:
-                            self.connected_pairs.append([i,j]) # store connected groups
-
-            elif method == "distance": 
-                index_overlap = np.where(np.linalg.norm(splist[:,3:] - sp1, ord=2, axis=-1) <= 1.5*radius)  # 2*distance_scale*radius
-                
-                if not np.any(index_overlap): # calculate their neighbors ...1)
-                    continue
-
-                neigbor_sp = splist[index_overlap] # calculate their neighbor ...2)
-                neigbor_sp = neigbor_sp[neigbor_sp[:,1] > i]
-                for j in neigbor_sp[:,1]:
-                    # two groups merge when their starting points distance is smaller than 1.5*radius
-                    self.connected_pairs.append([i,j]) # store connected groups
-
-            else:
-                raise ValueError(
-                    f"The method '{method}' does not exist,"
-                    f"please assign a new value")
+        self.connected_pairs = amalgamate(data, splist, radius, method)
 
         # self.check_labels = labels
         # reassign lalels, start from 0, since the cluster number not start with 0.
         
         # we employ an intutive and simple way to implement merging groups, resulting in a fast clustering
         self.merge_groups = merge_pairs(self.connected_pairs)
-
         maxid = max(labels) + 1
         
         # after this step, the connected pairs (groups) will be transformed into merged clusters, 
@@ -1612,109 +1562,12 @@ class classix:
 # *************************** <!-- the independent functions of checking overlap ***************************
 # *******************************  determine if two groups should be merged ********************************
 
-# import numpy as np
-from scipy.sparse import csr_matrix
-from scipy.special import betainc, gamma
-
-def check_if_overlap(starting_point, spo, radius, scale = 1):
-    """Check if two groups formed by aggregation overlap
-    """
-
-    if np.linalg.norm(starting_point - spo, ord=2, axis=-1) <= 2 * scale * radius:
-        return True
-    else:
-        return False
-
-    
-
-def cal_inter_density(starting_point, spo, radius, num):
-    """Calculate the density of intersection (lens)
-    """
-    in_volume = cal_inter_volume(starting_point, spo, radius)
-    return num / in_volume
-
-
-
-# deprecated (24/07/2021)
-# def cal_inter_area(starting_point, spo, radius):
-#     """Calculate the area of intersection (lens)
-#     """
-#     
-#     area = 2 * (radius**2) * np.arccos(dist / (2 * radius)) - dist * np.sqrt(4*radius**2 - dist**2) / 2
-#     return area    
-
-
-
-def cal_inter_volume(starting_point, spo, radius):
-    """
-    Returns the volume of the intersection of two spheres in n-dimensional space.
-    The radius of the two spheres is r and the distance of their centers is d.
-    For d=0 the function returns the volume of full sphere.
-    Reference: https://math.stackexchange.com/questions/162250/how-to-compute-the-volume-of-intersection-between-two-hyperspheres
-
-    """
-    
-    dim =starting_point.shape[0]
-    dist = np.linalg.norm(starting_point - spo, ord=2, axis=-1) # the distance between the two groups
-    
-    if dist > 2*radius:
-        return 0
-    c = dist / 2
-    return np.pi**(dim/2)/gamma(dim/2 + 1)*(radius**dim)*betainc((dim + 1)/2, 1/2, 1 - c**2/radius**2)
-
-
-
-# deprecated (24/07/2021)
-# def density(num, volume):
-#     ''' Calculate the density
-#     num: number of objects inside the cluster
-#     volume: the area of the cluster
-#     '''
-#     return num / volume
-    
-
-    
-def merge_pairs(pairs):
-    """Transform connected pairs to connected groups (list)"""
-    
-    len_p = len(pairs)
-    ulabels = np.full(len_p, -1, dtype=int)
-    labels = list()
-    maxid = 0
-    for i in range(len_p):
-        if ulabels[i] == -1:
-            sub = pairs[i]
-            ulabels[i] = maxid
-
-            for j in range(i+1, len_p):
-                com = pairs[j]
-                if check_if_intersect(sub, com):
-                    sub = sub + com
-                    if ulabels[j] == -1:
-                        ulabels[j] = maxid
-                    else:
-                        ulabels[ulabels == maxid] = ulabels[j]
-
-            maxid = maxid + 1
-
-    for i in np.unique(ulabels):
-        sub = list()
-        for j in np.where(ulabels == i)[0]:
-            sub = sub + pairs[int(j)]
-        labels.append(list(set(sub)))
-    return labels
-    
-
-
-def check_if_intersect(g1, g2):
-    """Check if two list have the same elements."""
-    return set(g1).intersection(g2) != set()
 
 
 
 def pairs_to_graph(pairs, num_nodes, sparse=True):
     """Transform the pairs represented by list into graph."""
-    from scipy.sparse import csr_matrix
+    # from scipy.sparse import csr_matrix
     graph = np.full((num_nodes, num_nodes), np.inf)
     for i in range(num_nodes):
         graph[i, i] = 0
